# SPDX-License-Identifier: Apache-2.0

from .onnx_operator_mixin import OnnxOperatorMixin


class OnnxSubGraphOperatorMixin(OnnxOperatorMixin):
    """
    :class:`OnnxOperatorMixin` for converters.
    """

    pass
